package tsp.csc4509.spool;

import static tsp.csc4509.common.Log.GEN;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;

import tsp.csc4509.common.Log;
import tsp.csc4509.tpnote.ProtocolMessage;

public class Spool {
	private static final String SPOOLNOTINITIALIZED = "L'instance spool n'a pas éte initilialisée";
	private static final String DIRNOTINSPOOL = "impossible de détruire un répertoire qui n'est pas dans les spools du service";
	private static final int BUFFERSIZE = 4096;
	public static final String FILEDIR = "file";

	private Path root;

	private static Spool instance = null;

	/**
	 * Constructeur privé: cette instance ne doit être créée qu'un seule fois par le
	 * client ou le serveur. La construction se fait par la méthode de classe
	 * initSpool() qui maintient cette propriété.
	 * 
	 * @param rootName répertoire où l'arborescence des fichiers est stockée.
	 * @throws IOException Exceptions d'entrées/sorties levées lors de la création
	 *                     de l'arborescence.
	 */
	private Spool(final String rootName) throws IOException {
		root = Paths.get(rootName);
		createRoot(rootName);

	}

	private void createRoot(final String rootName) throws IOException {
		if (Files.isDirectory(root)) {
			return;
		}
		Path newDir = Paths.get(rootName);
		Files.createDirectories(newDir);

	}

	/**
	 * Methode de classe qui construit l'objet s'il n'est pas déjà construit, et
	 * sinon retourne la référence de l'objet déjà construit.
	 * 
	 * @param rootName répertoire où l'arborescence des fichiers est stockée.
	 * @throws IOException Exceptions d'entrées/sorties levées lors de la création
	 *                     de l'arborescence.
	 */
	public static synchronized void initSpool(final String rootName) throws IOException {
		if (instance == null) {
			instance = new Spool(rootName);
		}
	}

	/**
	 * Méthode ce classe qui crée le repertoire pour mémoriser un fichier et les
	 * informations le concernant.
	 * 
	 * @param from     expéditeur du fichier.
	 * @param to       destinataire du fichier.
	 * @param fileName nom du fichier (juste le nom final, sans répertoire le
	 *                 précédent).
	 * @return Fichier ouvert en écriture sur le nouveau fichier à mémoriser.
	 * @throws Exception Exceptions levées lors de la création de l'arborescence.
	 */
//	public static synchronized FileOutputStream getNewOutputFile(final String from, final String to, final String fileName) throws Exception {
//		if (INSTANCE == null) throw new Exception(SPOOLNOTINITIALIZED);
//		Path path =  INSTANCE.createNewPath(from, to, fileName);
//		return INSTANCE.openOutputFile(path, fileName);
//	}

	public static synchronized Path initNewPath(final String from, final String to, final String fileName)
			throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return instance.createNewPath(from, to, fileName);
	}

	public static synchronized Path initNewPath(final Information information) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return instance.createNewPath(information.getFrom(), information.getTo(), information.getFileName());
	}

	public static synchronized Path initNewPath(final ProtocolMessage receiveMessage) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		Information information = new Information(receiveMessage.getFrom(), receiveMessage.getTo(),
				receiveMessage.getFileName());

		return instance.createNewPath(information.getFrom(), information.getTo(), information.getFileName());
	}

	public static synchronized FileInputStream getInputFile(final Path path) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return instance.openInputFile(path);
	}

	public static synchronized FileOutputStream getOutputFile(final Path path) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return instance.openOutputFile(path);
	}

	public static synchronized FileOutputStream getOutputStream(final ProtocolMessage receiveMessage) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		Path path = instance.createNewPath(receiveMessage.getFrom(), receiveMessage.getTo(),
				receiveMessage.getFileName());
		Information info = new Information(path);
		Path filePath = path.resolve(FILEDIR).resolve(info.getFileName());
		return new FileOutputStream(filePath.toString());

	}

	public static String getFrom(final Path path) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return new Information(path).getFrom();
	}

	public static String getTo(final Path path) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return new Information(path).getTo();
	}

	public static String getFileName(final Path path) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return new Information(path).getFileName();
	}

	public static Information getInformation(final Path path) throws Exception {
		return new Information(path);
	}

	public static String getPathFileName(final Path path) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return instance.root.resolve(FILEDIR).resolve(getFileName(path)).toString();
	}

	public static List<Path> getFilesListTo(final String to) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		return instance.listFileTo(to);

	}

	public static void deletePath(final Path path, final String spoolDir) throws Exception {
		if (instance == null)
			throw new Exception(SPOOLNOTINITIALIZED);
		String pathName = path.toString();
		// on ne veut détruire des répertoire que dans les spools!!
		if (pathName.indexOf(spoolDir) != 0) {
			throw new Exception(DIRNOTINSPOOL);
		}
		FileUtils.deleteDirectory(new File(pathName));

	}

//	private void load() throws IOException {
//		Pattern pattern = Pattern.compile("\\d+$");
//		try (DirectoryStream<Path> rootStream = Files.newDirectoryStream(root)) {
//			Iterator<Path> iterator = rootStream.iterator();
//			while(iterator.hasNext()) {
//				Path p = iterator.next();
//
//				String fileName = p.getFileName().toString();
//				Matcher matcher = pattern.matcher(fileName);
//				if (matcher.find()) {
//					int val = Integer.parseInt(fileName);
//					if (val > lastDir) {
//						lastDir = val;
//					}
//					if (Files.isDirectory(p)) {
//				//		loadDir(p);
//					}
//					System.out.println(p);
//				}
//			}
//		}
//	}

	private List<Path> listFileTo(final String to) throws IOException {
	//	GEN.info("listFileTo pour " + to);
		Pattern pattern = Pattern.compile("\\d+$");
		List<Path> listPath = new ArrayList<>();
		try (DirectoryStream<Path> rootStream = Files.newDirectoryStream(root)) {
			Iterator<Path> iterator = rootStream.iterator();
			// parcour du répertoire de la spool.
			while (iterator.hasNext()) {
				Path path = iterator.next();
		//		GEN.info("Exploration de " + path);
				// le nom du répertoire doit être un entier décimal?
				String fileName = path.getFileName().toString();
				Matcher matcher = pattern.matcher(fileName);

				if (matcher.find()) {
			//		GEN.info("Le nom du fichier est un entier");
					// ce répertoire doit contenir un fichier destiné à to.
					if (Files.isDirectory(path) && isFileTo(path, to)) {
				//		GEN.info("Le répertoire est ok");
						listPath.add(path);
					}
				}
			}
		}

		return listPath;
	}

	/**
	 * Indique si le chemin Path contient un fichier destiné à to.
	 * 
	 * @param path path d'un sous-répertoire présent dans le répertoire de la spool.
	 * @param to   nom du destinataire dont on recherche les fichiers.
	 * @return true si path contient un fichier destiné à to, false sinon.
	 */
	private boolean isFileTo(final Path path, final String to) {
		try {
			// chargement du fichier info du répertoire du spool.
			// lève une exception si ce fichier n'est pas présent ou pas conforme.
			Information info = new Information(path);

			// vérification du destintaire de ce fichier.
			if (!info.getTo().equals(to)) {
				return false;
			}

			// vérification de la présence d'un fichier.
			Path filePath = path.resolve(FILEDIR).resolve(info.getFileName());
			File file = new File(filePath.toString());
			if (!file.exists() || file.isDirectory()) {
				return false;
			}
			return true;

		} catch (Exception e) {
			return false;
		}
	}

	private int nextDir() throws IOException {
		int lastDir = 0;
		Pattern pattern = Pattern.compile("\\d+$");
		try (DirectoryStream<Path> rootStream = Files.newDirectoryStream(root)) {
			Iterator<Path> iterator = rootStream.iterator();
			while (iterator.hasNext()) {
				Path p = iterator.next();

				String fileName = p.getFileName().toString();
				Matcher matcher = pattern.matcher(fileName);
				if (matcher.find()) {
					int val = Integer.parseInt(fileName);
					if (val > lastDir) {
						lastDir = val;
					}
				}
			}
			return lastDir + 1;
		}
	}

	private synchronized Path createNewDir() throws IOException {
		int nextDir = nextDir();
		Path newDir = root.resolve("" + nextDir);
		Files.createDirectories(newDir);
		return newDir;
	}

	private void saveFile(ReadableByteChannel cin, WritableByteChannel cout) throws IOException {
		ByteBuffer buffer = ByteBuffer.allocate(BUFFERSIZE);
		while (cin.read(buffer) > 0) {
			buffer.flip();
			cout.write(buffer);
			buffer.clear();
		}
	}

	private synchronized FileChannel createNewFile(final Information information) throws IOException {
		//Log.GEN.info("create New File {}", information);
		Path newDir;
		newDir = createNewDir();

		Path fileDir = newDir.resolve(FILEDIR);
		Files.createDirectories(fileDir);
		Path filePath = fileDir.resolve(information.getFileName());
		Path infoFilePath = newDir.resolve(Information.INFORMATIONFILENAME);
		//Log.GEN.info("Info file {}", infoFilePath);
		try (PrintWriter writer = new PrintWriter(infoFilePath.toString());) {
			FileOutputStream fout = new FileOutputStream(filePath.toString());
			FileChannel fcout = fout.getChannel();
			writer.println(Information.FROM + information.getFrom());
			writer.println(Information.TO + information.getTo());
			writer.println(Information.FILENAME + information.getFileName());
		//	Log.GEN.info("fcput {}", fcout);
			return fcout;
		}
	}

	private synchronized Path createNewPath(final String from, final String to, final String fileName)
			throws IOException {
	//	Log.GEN.info("create Path");
		Path newDir;
		newDir = createNewDir();

		Path fileDir = newDir.resolve(FILEDIR);
		Files.createDirectories(fileDir);
		Path infoFilePath = newDir.resolve(Information.INFORMATIONFILENAME);
	//	Log.GEN.info("Info file {}", infoFilePath);
		try (PrintWriter writer = new PrintWriter(infoFilePath.toString());) {
			writer.println(Information.FROM + from);
			writer.println(Information.TO + to);
			writer.println(Information.FILENAME + fileName);
			return newDir;
		}
	}

	private synchronized FileInputStream openInputFile(final Path path) throws Exception {
		Information info = new Information(path);
		Path filePath = path.resolve(FILEDIR).resolve(info.getFileName());
		return new FileInputStream(filePath.toString());
	}

	private synchronized FileOutputStream openOutputFile(final Path path) throws Exception {
		Information info = new Information(path);
		Path filePath = path.resolve(FILEDIR).resolve(info.getFileName());
		return new FileOutputStream(filePath.toString());
	}

//	private int getLastDir() {
//		return lastDir;
//	}

//	private void loadInfo() {
//
//	}

}
