package tsp.csc4509.tpnote;

import static tsp.csc4509.common.Log.COMM;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import tsp.csc4509.spool.Information;
import tsp.csc4509.spool.Spool;
import tsp.csc4509.tcpnio.FullDuplexMsgWorker;
import tsp.csc4509.tcpnio.ReadMessageStatus;

/**
 * Classe du client.
 * 
 * @author
 *
 */
public class Client implements SenderReceiver {
	/**
	 * Taille des buffer.
	 */
	private static final int BUFFSIZE = 1024;

	/**
	 * répertoire ou seront stocké les spool des clients. Pour un client identifié
	 * par l'id userId, le répertoire de son sppol sera: CLIENTSSPOOLDIR/userId
	 */
	public static final String CLIENTSSPOOLDIR = "data/spool/";

	/**
	 * message d'erreur affiché en cas d'erreut de protocole.
	 */
	private static final String BADMESSAGE = "Bad message received from server.";

	/**
	 * Identifiant du client géré.
	 */
	private String userId;

	/**
	 * worker ouvert pour les échanges des messages du protocole.
	 */
	private FullDuplexMsgWorker worker;

	/**
	 * host du serveur. Mémorisé dans les attributs, car il est utilisé lors des
	 * connexions établies pour l'échange du contenu des fichiers.
	 */
	private String serverHost;

	/**
	 * Contructeur du client.
	 * 
	 * @param userId     Identifiant du client géré.
	 * @param serverHost host du serveur/
	 * @param serverPort port pour la connexion qui servira aux échanges des
	 *                   messages du protocole.
	 * @throws IOException toutes les exceptions qui peuvent se produire à
	 *                     l'établissement de la connexion et l'initialisation du
	 *                     spool du client.
	 */
	public Client(final String userId, final String serverHost, final int serverPort) throws IOException {
		super();
		this.serverHost = serverHost;
		worker = new FullDuplexMsgWorker(SocketChannel.open(new InetSocketAddress(serverHost, serverPort)));
		this.userId = userId;
		Spool.initSpool(CLIENTSSPOOLDIR + userId);
	}

	/**
	 * Envoi d'un message protocole PUT, et envoi du fichier désigné.
	 * 
	 * @param to       userId du destinataire du fichier/
	 * @param fileName nom du fichier à envoyer.
	 * @throws Exception Toutes les exceptions qui peuvent être levées pour la
	 *                   gestion de la connexion avec le serveur, et la lecture du
	 *                   fichier à envoyer.
	 */
	public void put(final String to, final String fileName) throws Exception {
		// on commence par ouvrir le fichier pour vérifier qu'il existe.
		FileInputStream fin = new FileInputStream(fileName);
		/*******************************************************************************/
		// Partie Protocole Interpreter: envoi PUT + réception RECEIVE
		// envoi du message de protcole PUT
		sendPutMessage(to, fileName);
		// réception du message de protcole SEND envoyé en réponse par le serveur.
		ProtocolMessage sendMessage = receiveProtocolMessage();
		if (sendMessage == null || sendMessage.getMessageType() != ProtocolMessageType.SEND) {
			fin.close();
			throw new Exception(BADMESSAGE + " " + sendMessage + " " + sendMessage.getMessageType());
		}
		/*******************************************************************************/
		// Partie Data Transfert Process: connexion au DTP serveur + envoi du fichier

		// établissement avec la socket serveur préparé par le serveur pour l'échange du
		// contenu
		// du fichier.
		SocketChannel rwChan = SocketChannel.open(new InetSocketAddress(serverHost, sendMessage.getPort()));
		FileChannel fc = fin.getChannel();
		sendFile(fc, rwChan);
		fin.close();
		fc.close();
		// fermeture obligatoire de la connexion pour que le serveur sache que le
		// fichier est totalement envoyé.
		rwChan.close();
	}

	/**
	 * Envoi d'un message protocole GET et réception de tous les fichiers concernés
	 * en parallèle dans une boucle gérant les réceptions en mode asynchrone.
	 * 
	 * @throws Exception Toutes les exceptions qui peuvent être levées pour la
	 *                   gestion des connexions avec le serveur, et l'écriture des
	 *                   fichiers à recevoir.
	 */
	public void get() throws Exception {
		/* #################### Debut corrigé Question 3 ##################### */
		// map pour lier les connexions aux serveur avec le fichier reçu.
		Map<SocketChannel, FileChannel> receiveFilesMap = new HashMap<>();
		// selector pour gérer les réceptions en parallèle.
		Selector selector = Selector.open();
		// on passer le worker en asynchrone pour qu'il puisse être géré par le
		// selector.
		worker.configureNonBlocking();
		SelectionKey key = worker.getChannel().register(selector, SelectionKey.OP_READ);
		COMM.info("Clef worker: {}", key);
		/*******************************************************************************/
		// Partie Protocole Interpreter: envoi du GET
		// envoi du message GET au serveur.
		sendGetMessage();
		// tant que l'on peut encore recevoir des nouveaux messages protocole RECEIVE et
		// tant que l'on continue de recevoir le contenu d'un des fichiers émis par le
		// serveur.
		while (worker.getChannel().isOpen() || receiveFilesMap.size() > 0) {
			COMM.info("Boucle selector");
			selector.select();
			Set<SelectionKey> readyKeys = selector.selectedKeys();
			Iterator<SelectionKey> it = readyKeys.iterator();
			while (it.hasNext()) {
				SelectionKey curKey = it.next();
				if (curKey.isReadable()) {
					/*******************************************************************************/
					// Partie Protocole Interpreter: réception SEND
					if (curKey.channel().equals(worker.getChannel())) {
						// gestion du canal pour les échanges des messages du protocol RECEIVE.
						COMM.info("Activité readable sur le canal des messages du protocole");
						ReadMessageStatus status = worker.readMessage();
						if (status == ReadMessageStatus.ChannelClosed) { // closed channel
							COMM.info("Fermeture du canal des messages du protocole");
							curKey.cancel();
							worker.close();
						}
						if (status == ReadMessageStatus.ReadDataCompleted) {
							switch (worker.getMessType()) {
							case MessType.PROTOCOLMESSAGETYPE:
								// reception d'un message protocole RECEIVE
								ProtocolMessage receiveMessage = (ProtocolMessage) worker.getData();
								COMM.info("Protocol Server: message {}", receiveMessage);
								if (receiveMessage == null
										|| receiveMessage.getMessageType() != ProtocolMessageType.RECEIVE) {
									break;
								}
								/*******************************************************************************/
								// Partie Data Transfert Process: connexion au DTP serveur
								// connexion avec le socker serveur prévu pour échanger le contenu du fichier.
								SocketChannel rwChan = SocketChannel
										.open(new InetSocketAddress(serverHost, receiveMessage.getPort()));
								rwChan.configureBlocking(false);
								key = rwChan.register(selector, SelectionKey.OP_READ);
								// ouverture du fichier en réception et mémorisation des informations dans la
								// Map receiveFilesMap
								FileOutputStream fout = Spool.getOutputStream(receiveMessage);
								FileChannel fcout = fout.getChannel();
								receiveFilesMap.put(rwChan, fcout);
								break;
							}
						}
					} else {
						/*******************************************************************************/
						// Partie Data Transfert Process: réception du fichier
						COMM.info("Activité readable sur un des canaux de réception des fichiers");
						ByteBuffer buffer = ByteBuffer.allocate(BUFFSIZE);
						SocketChannel rwChan = (SocketChannel) curKey.channel();
						if (rwChan.read(buffer) == -1) {
							// la connexion est fermé le fichier est totalement reçu.
							COMM.info("Data Server: closed channel detected");
							receiveFilesMap.get(rwChan).close();
							receiveFilesMap.remove(rwChan);
							rwChan.close();
							curKey.cancel();
						} else {
							// on ajoute les octets reçus au fichier.
							buffer.flip();
							receiveFilesMap.get(rwChan).write(buffer);
						}
					}
				}
			}
			readyKeys.clear();
		}
		/* #################### Fin corrigé Question 3 ##################### */
	}

	/**
	 * Envoi d'un message du protocole (PUT ou SEND) sur le worker du client. Cette méthode est
	 * appellée par les méthodes sendPutMessage() et sendGetMessage().
	 * @param message
	 *              le message du protocole à envoyer.
	 * @throws IOException
	 *              toutes les exceptions levées lors de l'écritures des messages.
	 */
	private void sendMessage(final ProtocolMessage message) throws IOException {
		worker.sendMsg(MessType.PROTOCOLMESSAGETYPE, message);
	}


	/**
	 * Envoi du message de protocole PUT sur le worker.
	 * @param to
	 *           nom de l'utilisateur destinataire du fichier.
	 * @param fileName
	 *           nom complet (répertoir inclus) du fichier à envoyer.
	 * @throws IOException
	 *            toutes les exceptions levées lors de l'écritures des messages.
	 */
	private void sendPutMessage(final String to, final String fileName) throws IOException {
		Information information = new Information(userId, to, fileName);
		ProtocolMessage putMessage = new ProtocolMessage(ProtocolMessageType.PUT, information);
		sendMessage(putMessage);
	}

	/**
	 * Envoi du message de protocole GET sur le worker.
	 * @throws IOException
	 *            toutes les exceptions levées lors de l'écritures des messages.
	 */
	private void sendGetMessage() throws IOException {
		COMM.info("Client::sendGetMessage()");
		Information information = new Information(userId, null, null);
		ProtocolMessage putMessage = new ProtocolMessage(ProtocolMessageType.GET, information);
		sendMessage(putMessage);
	}

	/**
	 * réception en mode synchrone des messages du protocole SEND. Ne peut pas être utilisée pour
	 * la réception en mode asynchrone des messages du protocole RECEIVE.
	 * @return
	 *         Le message SEND reçu. Le résultat null peut être interprété comme une fermeture de connexion du distant.
	 * @throws IOException
	 *         toutes les exceptions levées lors de la réception des messages.
	 */
	private ProtocolMessage receiveProtocolMessage() throws IOException {
		ReadMessageStatus status = ReadMessageStatus.ReadUnstarted;

		while ((status != ReadMessageStatus.ChannelClosed) && (status != ReadMessageStatus.ReadDataCompleted)) {
			status = worker.readMessage();
		}

		if (status == ReadMessageStatus.ReadDataCompleted) {
			Serializable obj = worker.getData();
			if (obj instanceof ProtocolMessage) {
				return (ProtocolMessage) obj;
			}
		}
		return null;

	}
}
